/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { RunOnceScheduler } from '../../../base/common/async';
import { CancellationToken, CancellationTokenSource } from '../../../base/common/cancellation';
import { DisposableStore, toDisposable } from '../../../base/common/lifecycle';
import { parseLinkedText } from '../../../base/common/linkedText';
import { LRUCache } from '../../../base/common/map';
import { assertType } from '../../../base/common/types';
import { URI } from '../../../base/common/uri';
import { DynamicCssRules } from '../../browser/editorDom';
import { registerEditorContribution } from '../../browser/editorExtensions';
import { EDITOR_FONT_DEFAULTS } from '../../common/config/editorOptions';
import { Range } from '../../common/core/range';
import { InlayHintKind, InlayHintsProviderRegistry } from '../../common/languages';
import { LanguageFeatureRequestDelays } from '../../common/languages/languageFeatureRegistry';
import { ModelDecorationInjectedTextOptions } from '../../common/model/textModel';
import { ITextModelService } from '../../common/services/resolverService';
import { ClickLinkGesture } from '../gotoSymbol/link/clickLinkGesture';
import { InlayHintsFragments } from './inlayHints';
import { CommandsRegistry } from '../../../platform/commands/common/commands';
import { IOpenerService } from '../../../platform/opener/common/opener';
import * as colors from '../../../platform/theme/common/colorRegistry';
import { themeColorFromId } from '../../../platform/theme/common/themeService';
const MAX_DECORATORS = 1500;
class InlayHintsCache {
    constructor() {
        this._entries = new LRUCache(50);
    }
    get(model) {
        const key = InlayHintsCache._key(model);
        return this._entries.get(key);
    }
    set(model, value) {
        const key = InlayHintsCache._key(model);
        this._entries.set(key, value);
    }
    static _key(model) {
        return `${model.uri.toString()}/${model.getVersionId()}`;
    }
}
export class InlayHintLabelPart {
    constructor(item, index, href) {
        this.item = item;
        this.index = index;
        this.href = href;
    }
}
let InlayHintsController = class InlayHintsController {
    constructor(_editor, _openerService) {
        this._editor = _editor;
        this._openerService = _openerService;
        this._disposables = new DisposableStore();
        this._sessionDisposables = new DisposableStore();
        this._getInlayHintsDelays = new LanguageFeatureRequestDelays(InlayHintsProviderRegistry, 25, 500);
        this._cache = new InlayHintsCache();
        this._decorationsMetadata = new Map();
        this._ruleFactory = new DynamicCssRules(this._editor);
        this._disposables.add(InlayHintsProviderRegistry.onDidChange(() => this._update()));
        this._disposables.add(_editor.onDidChangeModel(() => this._update()));
        this._disposables.add(_editor.onDidChangeModelLanguage(() => this._update()));
        this._disposables.add(_editor.onDidChangeConfiguration(e => {
            if (e.hasChanged(126 /* inlayHints */)) {
                this._update();
            }
        }));
        this._update();
    }
    dispose() {
        this._sessionDisposables.dispose();
        this._removeAllDecorations();
        this._disposables.dispose();
    }
    _update() {
        this._sessionDisposables.clear();
        this._removeAllDecorations();
        if (!this._editor.getOption(126 /* inlayHints */).enabled) {
            return;
        }
        const model = this._editor.getModel();
        if (!model || !InlayHintsProviderRegistry.has(model)) {
            return;
        }
        // iff possible, quickly update from cache
        const cached = this._cache.get(model);
        if (cached) {
            this._updateHintsDecorators([model.getFullModelRange()], cached);
        }
        const scheduler = new RunOnceScheduler(() => __awaiter(this, void 0, void 0, function* () {
            const t1 = Date.now();
            const cts = new CancellationTokenSource();
            this._sessionDisposables.add(toDisposable(() => cts.dispose(true)));
            const ranges = this._getHintsRanges();
            const inlayHints = yield InlayHintsFragments.create(model, ranges, cts.token);
            this._sessionDisposables.add(inlayHints);
            this._sessionDisposables.add(inlayHints.onDidReceiveProviderSignal(() => scheduler.schedule()));
            scheduler.delay = this._getInlayHintsDelays.update(model, Date.now() - t1);
            if (cts.token.isCancellationRequested) {
                return;
            }
            this._updateHintsDecorators(ranges, inlayHints.items);
            this._cacheHintsForFastRestore(model);
        }), this._getInlayHintsDelays.get(model));
        this._sessionDisposables.add(scheduler);
        // update inline hints when content or scroll position changes
        this._sessionDisposables.add(this._editor.onDidChangeModelContent(() => scheduler.schedule()));
        this._sessionDisposables.add(this._editor.onDidScrollChange(() => scheduler.schedule()));
        scheduler.schedule();
        // link gesture
        this._sessionDisposables.add(this._installLinkGesture());
    }
    _installLinkGesture() {
        let removeHighlight = () => { };
        const gesture = new ClickLinkGesture(this._editor);
        gesture.onMouseMoveOrRelevantKeyDown(e => {
            var _a, _b;
            const [mouseEvent] = e;
            if (mouseEvent.target.type !== 6 /* CONTENT_TEXT */ || typeof mouseEvent.target.detail !== 'object' || !mouseEvent.hasTriggerModifier) {
                removeHighlight();
                return;
            }
            const model = this._editor.getModel();
            const options = (_b = (_a = mouseEvent.target.detail) === null || _a === void 0 ? void 0 : _a.injectedText) === null || _b === void 0 ? void 0 : _b.options;
            if (options instanceof ModelDecorationInjectedTextOptions && options.attachedData instanceof InlayHintLabelPart && options.attachedData.href) {
                this._activeInlayHintPart = options.attachedData;
                const lineNumber = this._activeInlayHintPart.item.hint.position.lineNumber;
                const range = new Range(lineNumber, 1, lineNumber, model.getLineMaxColumn(lineNumber));
                const lineHints = new Set();
                for (let data of this._decorationsMetadata.values()) {
                    if (range.containsPosition(data.item.hint.position)) {
                        lineHints.add(data.item);
                    }
                }
                this._updateHintsDecorators([range], Array.from(lineHints));
                removeHighlight = () => {
                    this._activeInlayHintPart = undefined;
                    this._updateHintsDecorators([range], Array.from(lineHints));
                };
            }
        });
        gesture.onCancel(removeHighlight);
        gesture.onExecute(e => {
            var _a, _b;
            if (e.target.type !== 6 /* CONTENT_TEXT */ || typeof e.target.detail !== 'object' || !e.hasTriggerModifier) {
                return;
            }
            const options = (_b = (_a = e.target.detail) === null || _a === void 0 ? void 0 : _a.injectedText) === null || _b === void 0 ? void 0 : _b.options;
            if (options instanceof ModelDecorationInjectedTextOptions && options.attachedData instanceof InlayHintLabelPart && options.attachedData.href) {
                this._openerService.open(options.attachedData.href, { allowCommands: true, openToSide: e.hasSideBySideModifier });
            }
        });
        return gesture;
    }
    _cacheHintsForFastRestore(model) {
        const items = new Set();
        for (const [id, obj] of this._decorationsMetadata) {
            if (items.has(obj.item)) {
                // an inlay item can be rendered as multiple decorations
                // but they will all uses the same range
                continue;
            }
            items.add(obj.item);
            const range = model.getDecorationRange(id);
            if (range) {
                // update range with whatever the editor has tweaked it to
                obj.item.anchor.range = range;
            }
        }
        this._cache.set(model, Array.from(items));
    }
    _getHintsRanges() {
        const extra = 30;
        const model = this._editor.getModel();
        const visibleRanges = this._editor.getVisibleRangesPlusViewportAboveBelow();
        const result = [];
        for (const range of visibleRanges.sort(Range.compareRangesUsingStarts)) {
            const extendedRange = model.validateRange(new Range(range.startLineNumber - extra, range.startColumn, range.endLineNumber + extra, range.endColumn));
            if (result.length === 0 || !Range.areIntersectingOrTouching(result[result.length - 1], extendedRange)) {
                result.push(extendedRange);
            }
            else {
                result[result.length - 1] = Range.plusRange(result[result.length - 1], extendedRange);
            }
        }
        return result;
    }
    _updateHintsDecorators(ranges, items) {
        var _a, _b;
        const { fontSize, fontFamily } = this._getLayoutInfo();
        const newDecorationsData = [];
        const fontFamilyVar = '--code-editorInlayHintsFontFamily';
        this._editor.getContainerDomNode().style.setProperty(fontFamilyVar, fontFamily);
        for (const item of items) {
            // text w/ links
            const { nodes } = parseLinkedText(item.hint.label);
            const marginBefore = item.hint.whitespaceBefore ? (fontSize / 3) | 0 : 0;
            const marginAfter = item.hint.whitespaceAfter ? (fontSize / 3) | 0 : 0;
            for (let i = 0; i < nodes.length; i++) {
                const node = nodes[i];
                const isFirst = i === 0;
                const isLast = i === nodes.length - 1;
                const isLink = typeof node === 'object';
                const cssProperties = {
                    fontSize: `${fontSize}px`,
                    fontFamily: `var(${fontFamilyVar}), ${EDITOR_FONT_DEFAULTS.fontFamily}`,
                    verticalAlign: 'middle',
                };
                this._fillInColors(cssProperties, item.hint);
                if (isLink) {
                    cssProperties.textDecoration = 'underline';
                    if (((_a = this._activeInlayHintPart) === null || _a === void 0 ? void 0 : _a.item) === item && this._activeInlayHintPart.index === i && this._activeInlayHintPart.href === node.href) {
                        // active link!
                        cssProperties.cursor = 'pointer';
                        cssProperties.color = themeColorFromId(colors.editorActiveLinkForeground);
                    }
                }
                if (isFirst && isLast) {
                    // only element
                    cssProperties.margin = `0px ${marginAfter}px 0px ${marginBefore}px`;
                    cssProperties.padding = `1px ${Math.max(1, fontSize / 4) | 0}px`;
                    cssProperties.borderRadius = `${(fontSize / 4) | 0}px`;
                }
                else if (isFirst) {
                    // first element
                    cssProperties.margin = `0px 0 0 ${marginAfter}px`;
                    cssProperties.padding = `1px 0 0 ${Math.max(1, fontSize / 4) | 0}px`;
                    cssProperties.borderRadius = `${(fontSize / 4) | 0}px 0 0 ${(fontSize / 4) | 0}px`;
                }
                else if (isLast) {
                    // last element
                    cssProperties.margin = `0px ${marginAfter}px 0 0`;
                    cssProperties.padding = `1px ${Math.max(1, fontSize / 4) | 0}px 0 0`;
                    cssProperties.borderRadius = `0 ${(fontSize / 4) | 0}px ${(fontSize / 4) | 0}px 0`;
                }
                else {
                    cssProperties.padding = `1px 0 1px 0`;
                }
                const classNameRef = this._ruleFactory.createClassNameRef(cssProperties);
                newDecorationsData.push({
                    item,
                    classNameRef,
                    decoration: {
                        range: item.anchor.range,
                        options: {
                            [item.anchor.direction]: {
                                content: fixSpace(isLink ? node.label : node),
                                inlineClassNameAffectsLetterSpacing: true,
                                inlineClassName: classNameRef.className,
                                attachedData: new InlayHintLabelPart(item, i, isLink ? node.href : undefined)
                            },
                            description: 'InlayHint',
                            showIfCollapsed: !item.anchor.usesWordRange,
                            stickiness: 0 /* AlwaysGrowsWhenTypingAtEdges */
                        }
                    },
                });
            }
            if (newDecorationsData.length > MAX_DECORATORS) {
                break;
            }
        }
        // collect all decoration ids that are affected by the ranges
        // and only update those decorations
        const decorationIdsToReplace = [];
        for (const range of ranges) {
            for (const { id } of (_b = this._editor.getDecorationsInRange(range)) !== null && _b !== void 0 ? _b : []) {
                const metadata = this._decorationsMetadata.get(id);
                if (metadata) {
                    decorationIdsToReplace.push(id);
                    metadata.classNameRef.dispose();
                    this._decorationsMetadata.delete(id);
                }
            }
        }
        const newDecorationIds = this._editor.deltaDecorations(decorationIdsToReplace, newDecorationsData.map(d => d.decoration));
        for (let i = 0; i < newDecorationIds.length; i++) {
            const data = newDecorationsData[i];
            this._decorationsMetadata.set(newDecorationIds[i], { item: data.item, classNameRef: data.classNameRef });
        }
    }
    _fillInColors(props, hint) {
        if (hint.kind === InlayHintKind.Parameter) {
            props.backgroundColor = themeColorFromId(colors.editorInlayHintParameterBackground);
            props.color = themeColorFromId(colors.editorInlayHintParameterForeground);
        }
        else if (hint.kind === InlayHintKind.Type) {
            props.backgroundColor = themeColorFromId(colors.editorInlayHintTypeBackground);
            props.color = themeColorFromId(colors.editorInlayHintTypeForeground);
        }
        else {
            props.backgroundColor = themeColorFromId(colors.editorInlayHintBackground);
            props.color = themeColorFromId(colors.editorInlayHintForeground);
        }
    }
    _getLayoutInfo() {
        const options = this._editor.getOption(126 /* inlayHints */);
        const editorFontSize = this._editor.getOption(45 /* fontSize */);
        let fontSize = options.fontSize;
        if (!fontSize || fontSize < 5 || fontSize > editorFontSize) {
            fontSize = (editorFontSize * .9) | 0;
        }
        const fontFamily = options.fontFamily || this._editor.getOption(42 /* fontFamily */);
        return { fontSize, fontFamily };
    }
    _removeAllDecorations() {
        this._editor.deltaDecorations(Array.from(this._decorationsMetadata.keys()), []);
        for (let obj of this._decorationsMetadata.values()) {
            obj.classNameRef.dispose();
        }
        this._decorationsMetadata.clear();
    }
};
InlayHintsController.ID = 'editor.contrib.InlayHints';
InlayHintsController = __decorate([
    __param(1, IOpenerService)
], InlayHintsController);
export { InlayHintsController };
// Prevents the view from potentially visible whitespace
function fixSpace(str) {
    const noBreakWhitespace = '\xa0';
    return str.replace(/[ \t]/g, noBreakWhitespace);
}
registerEditorContribution(InlayHintsController.ID, InlayHintsController);
CommandsRegistry.registerCommand('_executeInlayHintProvider', (accessor, ...args) => __awaiter(void 0, void 0, void 0, function* () {
    const [uri, range] = args;
    assertType(URI.isUri(uri));
    assertType(Range.isIRange(range));
    const ref = yield accessor.get(ITextModelService).createModelReference(uri);
    try {
        const model = yield InlayHintsFragments.create(ref.object.textEditorModel, [Range.lift(range)], CancellationToken.None);
        const result = model.items.map(i => i.hint);
        setTimeout(() => model.dispose(), 0); // dispose after sending to ext host
        return result;
    }
    finally {
        ref.dispose();
    }
}));
